import logging
import threading
import time
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime, timezone
from pathlib import Path

import urllib3
from minio import Minio
from minio.error import S3Error

from pymilvus.bulk_writer.constants import ConnectType
from pymilvus.bulk_writer.endpoint_resolver import EndpointResolver
from pymilvus.bulk_writer.file_utils import FileUtils
from pymilvus.bulk_writer.stage_restful import apply_stage

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class StageFileManager:
    def __init__(
        self,
        cloud_endpoint: str,
        api_key: str,
        stage_name: str,
        connect_type: ConnectType = ConnectType.AUTO,
    ):
        """
        private preview feature. Please submit a request and contact us if you need it.

        Args:
            cloud_endpoint (str): The fixed cloud endpoint URL.
                - For international regions: https://api.cloud.zilliz.com
                - For regions in China: https://api.cloud.zilliz.com.cn
            api_key (str): The API key associated with your organization
            stage_name (str): The name of the Stage.
            connect_type: Current value is mainly for Aliyun OSS buckets, default is Auto.
             - Default case, if the OSS bucket is reachable via the internal endpoint,
               the internal endpoint will be used
             - otherwise, the public endpoint will be used.
             - You can also force the use of either the internal or public endpoint.
        """
        self.cloud_endpoint = cloud_endpoint
        self.api_key = api_key
        self.stage_name = stage_name
        self.connect_type = connect_type
        self.local_file_paths = []
        self.total_bytes = 0
        self.stage_info = {}
        self._client = None

    def _convert_dir_path(self, input_path: str):
        if not input_path or input_path == "/":
            return ""
        if input_path.endswith("/"):
            return input_path
        return input_path + "/"

    def _refresh_stage_and_client(self, path: str):
        logger.info("refreshing stage info...")
        response = apply_stage(self.cloud_endpoint, self.api_key, self.stage_name, path)
        self.stage_info = response.json()["data"]
        logger.info("stage info refreshed.")

        creds = self.stage_info["credentials"]
        http_client = urllib3.PoolManager(maxsize=100)

        endpoint = EndpointResolver.resolve_endpoint(
            self.stage_info["endpoint"],
            self.stage_info["cloud"],
            self.stage_info["region"],
            self.connect_type,
        )
        self._client = Minio(
            endpoint=endpoint,
            access_key=creds["tmpAK"],
            secret_key=creds["tmpSK"],
            session_token=creds["sessionToken"],
            region=self.stage_info["region"],
            secure=True,
            http_client=http_client,
        )
        logger.info("storage client refreshed")

    def _validate_size(self):
        file_size_total = self.total_bytes
        file_size_limit = self.stage_info["condition"]["maxContentLength"]
        if file_size_total > file_size_limit:
            error_message = (
                f"localFileTotalSize {file_size_total} exceeds "
                f"the maximum contentLength limit {file_size_limit} defined in the condition."
                f"If you want to upload larger files, please contact us to lift the restriction."
            )
            raise ValueError(error_message)

    def upload_file_to_stage(self, source_file_path: str, target_stage_path: str):
        """
        uploads a local file or directory to the specified path within the Stage.

        Args:
            source_file_path: the source local file or directory path
            target_stage_path: the target directory path in the Stage
        Raises:
            Exception: If an error occurs during the upload process.
        """

        self.local_file_paths, self.total_bytes = FileUtils.process_local_path(source_file_path)
        stage_path = self._convert_dir_path(target_stage_path)
        self._refresh_stage_and_client(stage_path)
        self._validate_size()

        file_count = len(self.local_file_paths)
        logger.info(
            f"begin to upload file to stage, localDirOrFilePath:{source_file_path}, fileCount:{file_count} to stageName:{self.stage_name}, stagePath:{stage_path}"
        )
        start_time = time.time()

        uploaded_bytes = 0
        uploaded_count = 0
        root_path = Path(source_file_path).resolve()
        uploaded_bytes_lock = threading.Lock()

        def _upload_task(file_path: str, root_path: Path, stage_path: str):
            nonlocal uploaded_bytes
            nonlocal uploaded_count
            path_obj = Path(file_path).resolve()
            if root_path.is_file():
                relative_path = path_obj.name
            else:
                relative_path = path_obj.relative_to(root_path).as_posix()

            stage_prefix = f"{self.stage_info['stagePrefix']}"
            file_start_time = time.time()
            try:
                size = Path(file_path).stat().st_size
                logger.info(f"uploading file, fileName:{file_path}, size:{size} bytes")
                remote_file_path = stage_prefix + stage_path + relative_path
                self._put_object(file_path, remote_file_path, stage_path)
                with uploaded_bytes_lock:
                    uploaded_bytes += size
                    uploaded_count += 1
                percent = uploaded_bytes / self.total_bytes * 100
                elapsed = time.time() - file_start_time
                logger.info(
                    f"Uploaded file, {uploaded_count}/{file_count}: {file_path}, elapsed:{elapsed} s, {uploaded_bytes}/{self.total_bytes} bytes, progress: {percent:.2f}%"
                )
            except S3Error as e:
                logger.error(f"Failed to upload {file_path}: {e!s}")
                raise

        with ThreadPoolExecutor(max_workers=20) as executor:
            futures = []
            for _, file_path in enumerate(self.local_file_paths):
                futures.append(executor.submit(_upload_task, file_path, root_path, stage_path))
            for f in futures:
                f.result()  # wait for all

        total_elapsed = time.time() - start_time
        logger.info(
            f"All files in {source_file_path} uploaded to stage, "
            f"stageName:{self.stage_info['stageName']}, stagePath: {stage_path}, "
            f"totalFileCount:{file_count}, totalFileSize:{self.total_bytes}, cost time:{total_elapsed}s"
        )
        return {"stageName": self.stage_info["stageName"], "path": stage_path}

    def _put_object(self, file_path: str, remote_file_path: str, stage_path: str):
        expire_time_str = self.stage_info["credentials"]["expireTime"]
        expire_time = datetime.fromisoformat(expire_time_str.replace("Z", "+00:00"))

        now = datetime.now(timezone.utc)
        if now > expire_time:
            self._refresh_stage_and_client(stage_path)

        self._upload_with_retry(file_path, remote_file_path, stage_path)

    def _upload_with_retry(
        self, file_path: str, object_name: str, stage_path: str, max_retries: int = 5
    ):
        attempt = 0
        while attempt < max_retries:
            try:
                self._client.fput_object(
                    bucket_name=self.stage_info["bucketName"],
                    object_name=object_name,
                    file_path=file_path,
                )
                break
            except Exception as e:
                attempt += 1
                logger.warning(f"Attempt {attempt} failed to upload {file_path}: {e}")
                self._refresh_stage_and_client(stage_path)

                if attempt == max_retries:
                    error_message = f"Upload failed after {max_retries} attempts"
                    raise RuntimeError(error_message) from e

                time.sleep(5)



import asyncio

from pymilvus import AsyncMilvusClient, DataType

# Super user credentials for admin operations
super_user = ""
super_password = ""

fmt = "\n=== {:30} ===\n"
collection_name = "test_privilege_collection"
test_user = "test_privilege_user"
test_role_v1 = "test_role_v1"
test_role_v2 = "test_role_v2"
test_db = "test_privilege_database"
privilege_group_name = "test_privilege_group"

# Test data for privilege operations
db_rw_privileges = [
    {"object_type": "Collection", "privilege": "Insert", "object_name": collection_name},
    {"object_type": "Collection", "privilege": "Delete", "object_name": collection_name},
    {"object_type": "Collection", "privilege": "Search", "object_name": collection_name},
    {"object_type": "Collection", "privilege": "Query", "object_name": collection_name},
]

db_ro_privileges = [
    {"object_type": "Collection", "privilege": "Search", "object_name": collection_name},
    {"object_type": "Collection", "privilege": "Query", "object_name": collection_name},
]


async def create_resources(client: AsyncMilvusClient):
    print(fmt.format("Creating Resources"))

    print("Creating database...")
    await client.create_database(test_db)
    print(f"Database {test_db} created")

    print("Creating user and roles...")
    await client.create_user(test_user, "password123")
    await client.create_role(test_role_v1)
    await client.create_role(test_role_v2)
    print(f"User {test_user} and roles {test_role_v1}, {test_role_v2} created")

    print("Creating privilege group...")
    await client.create_privilege_group(privilege_group_name)
    print(f"Privilege group {privilege_group_name} created")

    print("Adding privileges to group...")
    await client.add_privileges_to_group(privilege_group_name, ["Search", "Query", "Insert"])
    print("Privileges added to group")

    print("Creating collection...")
    schema = client.create_schema(
        auto_id=True, description="Test privilege collection", enable_dynamic_field=True
    )
    schema.add_field("id", DataType.INT64, is_primary=True, auto_id=True)
    schema.add_field("vector", DataType.FLOAT_VECTOR, dim=128)
    schema.add_field("text", DataType.VARCHAR, max_length=512)

    await client.create_collection(collection_name, schema=schema, using_database=test_db)
    print(f"Collection {collection_name} created in database {test_db}")

    return True


async def test_grant_revoke_privilege(client: AsyncMilvusClient):
    print(fmt.format("Testing grant_privilege / revoke_privilege"))

    # Test grant_privilege for different privileges
    print("Testing grant_privilege...")
    for privilege_item in db_rw_privileges:
        await client.grant_privilege(
            role_name=test_role_v1,
            object_type=privilege_item["object_type"],
            privilege=privilege_item["privilege"],
            object_name=privilege_item["object_name"],
            db_name=test_db,
        )
        print(
            f"Granted {privilege_item['privilege']} on {privilege_item['object_name']} to {test_role_v1}"
        )

    # Check role privileges
    print(f"\nChecking role privileges for {test_role_v1}...")
    role_info = await client.describe_role(test_role_v1)
    print(f"Role {test_role_v1} privileges: {role_info}")

    # Grant role to user
    print(f"\nGranting role {test_role_v1} to user {test_user}...")
    await client.grant_role(test_user, test_role_v1)
    print("Role granted to user")

    # Check user info
    user_info = await client.describe_user(test_user)
    print(f"User {test_user} info: {user_info}")

    # Test revoke_privilege
    print("\nTesting revoke_privilege...")
    for privilege_item in db_ro_privileges:
        await client.revoke_privilege(
            role_name=test_role_v1,
            object_type=privilege_item["object_type"],
            privilege=privilege_item["privilege"],
            object_name=privilege_item["object_name"],
            db_name=test_db,
        )
        print(
            f"Revoked {privilege_item['privilege']} on {privilege_item['object_name']} from {test_role_v1}"
        )

    # Check role privileges after revoke
    print(f"\nChecking role privileges after revoke for {test_role_v1}...")
    role_info = await client.describe_role(test_role_v1)
    print(f"Role {test_role_v1} privileges after revoke: {role_info}")

    # Revoke role from user
    print(f"\nRevoking role {test_role_v1} from user {test_user}...")
    await client.revoke_role(test_user, test_role_v1)
    print("Role revoked from user")


async def test_grant_revoke_privilege_v2(client: AsyncMilvusClient):
    print(fmt.format("Testing grant_privilege_v2 / revoke_privilege_v2"))

    # Test grant_privilege_v2 with custom privilege group
    print("Testing grant_privilege_v2 with custom privilege group...")
    await client.grant_privilege_v2(
        role_name=test_role_v2, privilege=privilege_group_name, collection_name="*", db_name=test_db
    )
    print(
        f"Granted privilege group {privilege_group_name} on all collections in {test_db} to {test_role_v2}"
    )

    # Test grant_privilege_v2 with built-in privilege groups
    print("\nTesting grant_privilege_v2 with built-in privilege groups...")

    # Grant collection-level privileges
    await client.grant_privilege_v2(
        role_name=test_role_v2,
        privilege="CollectionReadWrite",
        collection_name=collection_name,
        db_name=test_db,
    )
    print(f"Granted CollectionReadWrite on {collection_name} to {test_role_v2}")

    # Grant database-level privileges
    await client.grant_privilege_v2(
        role_name=test_role_v2, privilege="DatabaseReadOnly", collection_name="*", db_name=test_db
    )
    print(f"Granted DatabaseReadOnly on database {test_db} to {test_role_v2}")

    # Check role privileges
    print(f"\nChecking role privileges for {test_role_v2}...")
    role_info = await client.describe_role(test_role_v2)
    print(f"Role {test_role_v2} privileges: {role_info}")

    # Grant role to user
    print(f"\nGranting role {test_role_v2} to user {test_user}...")
    await client.grant_role(test_user, test_role_v2)
    print("Role granted to user")

    # Check user info
    user_info = await client.describe_user(test_user)
    print(f"User {test_user} info: {user_info}")

    # Test revoke_privilege_v2
    print("\nTesting revoke_privilege_v2...")

    # Revoke custom privilege group
    await client.revoke_privilege_v2(
        role_name=test_role_v2, privilege=privilege_group_name, collection_name="*", db_name=test_db
    )
    print(f"Revoked privilege group {privilege_group_name} from {test_role_v2}")

    # Revoke built-in privilege group
    await client.revoke_privilege_v2(
        role_name=test_role_v2, privilege="DatabaseReadOnly", collection_name="*", db_name=test_db
    )
    print(f"Revoked DatabaseReadOnly from {test_role_v2}")

    # Revoke collection-level privilege
    await client.revoke_privilege_v2(
        role_name=test_role_v2,
        privilege="CollectionReadWrite",
        collection_name=collection_name,
        db_name=test_db,
    )
    print(f"Revoked CollectionReadWrite from {test_role_v2}")

    # Check role privileges after revoke
    print(f"\nChecking role privileges after revoke for {test_role_v2}...")
    role_info = await client.describe_role(test_role_v2)
    print(f"Role {test_role_v2} privileges after revoke: {role_info}")

    # Revoke role from user
    print(f"\nRevoking role {test_role_v2} from user {test_user}...")
    await client.revoke_role(test_user, test_role_v2)
    print("Role revoked from user")


async def main():
    client = AsyncMilvusClient("http://localhost:19530", user=super_user, password=super_password)

    await create_resources(client)
    await test_grant_revoke_privilege(client)
    await test_grant_revoke_privilege_v2(client)
    print(fmt.format("All Privilege Tests Completed Successfully"))
    await client.close()


if __name__ == "__main__":
    asyncio.run(main())



from pathlib import Path
from typing import List, Tuple


class FileUtils:
    @staticmethod
    def process_local_path(local_path: str) -> Tuple[List[str], int]:
        """
        Get all file paths under the local_path, recursively.
        If local_path is a file, return it directly.
        Returns a tuple: (list of file paths, total size in bytes)
        """
        if not Path(local_path).exists():
            error_message = f"Path does not exist: {local_path}"
            raise ValueError(error_message)

        path = Path(local_path)
        if path.is_file():
            return [local_path], path.stat().st_size
        if path.is_dir():
            return FileUtils.find_files_recursively(local_path)

        return [], 0

    @staticmethod
    def find_files_recursively(folder_path: str) -> Tuple[List[str], int]:
        """
        Recursively finds all files under a folder and calculates total size.
        """
        result: List[str] = []
        total_size = 0

        for file_path in Path(folder_path).rglob("*"):
            if file_path.is_file():
                resolved_path = file_path.resolve()
                result.append(str(resolved_path))
                total_size += resolved_path.stat().st_size

        return result, total_size



import time
import numpy as np
from pymilvus import (
    MilvusClient,
)

fmt = "\n=== {:30} ===\n"
dim = 8
collection_name = "hello_client_cost"
milvus_client = MilvusClient("http://localhost:19530")

has_collection = milvus_client.has_collection(collection_name, timeout=5)
if has_collection:
    milvus_client.drop_collection(collection_name)
milvus_client.create_collection(collection_name, dim, consistency_level="Strong", metric_type="L2")

print(fmt.format("    all collections    "))
print(milvus_client.list_collections())

print(fmt.format(f"schema of collection {collection_name}"))
print(milvus_client.describe_collection(collection_name))

rng = np.random.default_rng(seed=19530)
rows = [
        {"id": 1, "vector": rng.random((1, dim))[0], "a": 100},
        {"id": 2, "vector": rng.random((1, dim))[0], "b": 200},
        {"id": 3, "vector": rng.random((1, dim))[0], "c": 300},
        {"id": 4, "vector": rng.random((1, dim))[0], "d": 400},
        {"id": 5, "vector": rng.random((1, dim))[0], "e": 500},
        {"id": 6, "vector": rng.random((1, dim))[0], "f": 600},
]

print(fmt.format("Start inserting entities"))
insert_result = milvus_client.insert(collection_name, rows, progress_bar=True)
print(fmt.format("Inserting entities done"))
# OUTPUT:
# insert result: {'insert_count': 6, 'ids': [1, 2, 3, 4, 5, 6], 'cost': '1'};
# insert cost: 1
print(f"insert result: {insert_result};\ninsert cost: {insert_result['cost']}")

print(fmt.format("Start query by specifying primary keys"))
query_results = milvus_client.query(collection_name, ids=[2])
# OUTPUT:
# query result: data: ["{'id': 2, 'vector': [0.9007387, 0.44944635, 0.18477614, 0.42930314, 0.40345728, 0.3957196, 0.6963897, 0.24356908], 'b': 200}"], extra_info: {'cost': '21'}
# query cost: 21
print(f"query result: {query_results}\nquery cost: {query_results.extra['cost']}")

upsert_ret = milvus_client.upsert(collection_name, {"id": 2 , "vector": rng.random((1, dim))[0], "g": 100})
# OUTPUT:
# upsert result: {'upsert_count': 1, 'cost': '2'}
# upsert cost: 2
print(f"upsert result: {upsert_ret}\nupsert cost: {upsert_ret['cost']}")

print(fmt.format("Start query by specifying primary keys"))
query_results = milvus_client.query(collection_name, ids=[2])
print(f"query result: {query_results}\nquery cost: {query_results.extra['cost']}")

print(f"start to delete by specifying filter in collection {collection_name}")
delete_result = milvus_client.delete(collection_name, ids=[6])
# OUTPUT:
# delete result: {'delete_count': 1, 'cost': '1'}
# delete cost: 1
print(f"delete result: {delete_result}\ndelete cost: {delete_result['cost']}")

rng = np.random.default_rng(seed=19530)
vectors_to_search = rng.random((1, dim))

print(fmt.format(f"Start search with retrieve serveral fields."))
result = milvus_client.search(collection_name, vectors_to_search, limit=3, output_fields=["pk", "a", "b"])
print(f"search result: {result}\nsearch cost: {result.extra['cost']}")

milvus_client.drop_collection(collection_name)



import time

import grpc
import pytest
from pymilvus.decorators import IGNORE_RETRY_CODES, retry_on_rpc_failure
from pymilvus.exceptions import ErrorCode, MilvusException
from pymilvus.grpc_gen import common_pb2


class TestDecorators:
    def mock_failure(self, code: grpc.StatusCode):
        if code in IGNORE_RETRY_CODES:
            raise MockUnRetriableError(code)
        if code == MockUnavailableError().code():
            raise MockUnavailableError
        if code == MockDeadlineExceededError().code():
            raise MockDeadlineExceededError

    def mock_milvus_exception(self, code: ErrorCode):
        if code == ErrorCode.FORCE_DENY:
            raise MockForceDenyError
        if code == ErrorCode.RATE_LIMIT:
            raise MockRateLimitError
        raise MilvusException(ErrorCode.UNEXPECTED_ERROR, "unexpected error")

    @pytest.mark.parametrize("times", [0, 1, 2, 3])
    def test_retry_decorators_unavailable(self, times):
        self.count_test_retry_decorators_Unavailable = 0

        @retry_on_rpc_failure(retry_times=times)
        def test_api(self, code):
            self.count_test_retry_decorators_Unavailable += 1
            self.mock_failure(code)

        with pytest.raises(MilvusException, match="unavailable"):
            test_api(self, grpc.StatusCode.UNAVAILABLE)

        # the first execute + retry times
        assert self.count_test_retry_decorators_Unavailable == times + 1

    def test_retry_decorators_timeout(self):
        self.count_test_retry_decorators_timeout = 0

        @retry_on_rpc_failure()
        def test_api(self, code, timeout=None):
            self.count_test_retry_decorators_timeout += 1
            time.sleep(1)
            self.mock_failure(code)

        with pytest.raises(MilvusException):
            test_api(self, grpc.StatusCode.UNAVAILABLE, timeout=1)

        assert self.count_test_retry_decorators_timeout == 1

    @pytest.mark.skip("Do not open this unless you have loads of time, get some coffee and wait")
    def test_retry_decorators_default_behaviour(self):
        self.test_retry_decorators_default_retry_times = 0

        @retry_on_rpc_failure()
        def test_api(self, code):
            self.test_retry_decorators_default_retry_times += 1
            self.mock_failure(code)

        with pytest.raises(MilvusException):
            test_api(self, grpc.StatusCode.UNAVAILABLE)

        assert self.test_retry_decorators_default_retry_times == 7 + 1

    def test_retry_decorators_force_deny(self):
        self.execute_times = 0

        @retry_on_rpc_failure()
        def test_api(self, code):
            self.execute_times += 1
            self.mock_milvus_exception(code)

        with pytest.raises(MilvusException, match="force deny"):
            test_api(self, ErrorCode.FORCE_DENY)

        # the first execute + 0 retry times
        assert self.execute_times == 1

    def test_retry_decorators_set_retry_times(self):
        self.count_retry_times = 0

        @retry_on_rpc_failure()
        def test_api(self, code, retry_on_rate_limit, **kwargs):
            self.count_retry_times += 1
            self.mock_milvus_exception(code)

        with pytest.raises(MilvusException):
            test_api(self, ErrorCode.RATE_LIMIT, retry_on_rate_limit=True, retry_times=3)

        # the first execute + 0 retry times
        assert self.count_retry_times == 3 + 1

    @pytest.mark.parametrize("times", [0, 1, 2, 3])
    def test_retry_decorators_rate_limit_without_retry(self, times):
        self.count_test_retry_decorators_force_deny = 0

        @retry_on_rpc_failure(retry_times=times)
        def test_api(self, code, retry_on_rate_limit):
            self.count_test_retry_decorators_force_deny += 1
            self.mock_milvus_exception(code)

        with pytest.raises(MilvusException, match="rate limit"):
            test_api(self, ErrorCode.RATE_LIMIT, retry_on_rate_limit=False)

        # the first execute + 0 retry times
        assert self.count_test_retry_decorators_force_deny == 1

    @pytest.mark.parametrize("times", [0, 1, 2, 3])
    def test_retry_decorators_rate_limit_with_retry(self, times):
        self.count_test_retry_decorators_force_deny = 0

        @retry_on_rpc_failure(retry_times=times)
        def test_api(self, code, retry_on_rate_limit):
            self.count_test_retry_decorators_force_deny += 1
            self.mock_milvus_exception(code)

        with pytest.raises(MilvusException, match="rate limit"):
            test_api(self, ErrorCode.RATE_LIMIT, retry_on_rate_limit=True)

        # the first execute + retry times
        assert self.count_test_retry_decorators_force_deny == times + 1

    @pytest.mark.parametrize("code", IGNORE_RETRY_CODES)
    def test_donot_retry_codes(self, code):
        self.count_test_donot_retry = 0

        @retry_on_rpc_failure()
        def test_api(self, code):
            self.count_test_donot_retry += 1
            self.mock_failure(code)

        with pytest.raises(grpc.RpcError):
            test_api(self, code)

        # no retry
        assert self.count_test_donot_retry == 1


class MockUnavailableError(grpc.RpcError):
    def code(self):
        return grpc.StatusCode.UNAVAILABLE

    def details(self):
        return "details of unavailable"


class MockDeadlineExceededError(grpc.RpcError):
    def code(self):
        return grpc.StatusCode.DEADLINE_EXCEEDED

    def details(self):
        return "details of deadline exceeded"


class MockForceDenyError(MilvusException):
    def __init__(self, code=ErrorCode.FORCE_DENY, message="force deny"):
        super(MilvusException, self).__init__(message)
        self._code = code
        self._message = message
        self._compatible_code = common_pb2.ForceDeny

    @property
    def code(self):
        return self._code

    @property
    def message(self):
        return self._message

    @property
    def compatible_code(self):
        return self._compatible_code


class MockRateLimitError(MilvusException):
    def __init__(self, code=ErrorCode.RATE_LIMIT, message="rate limit"):
        super(MilvusException, self).__init__(message)
        self._code = code
        self._message = message
        self._compatible_code = common_pb2.RateLimit

    @property
    def code(self):
        return self._code

    @property
    def message(self):
        return self._message

    @property
    def compatible_code(self):
        return self._compatible_code

class MockUnRetriableError(grpc.RpcError):
    def __init__(self, code: grpc.StatusCode, message="unretriable error"):
        self._code = code

    def code(self):
        return self._code

    def details(self):
        return "details of unretriable error"



import http
import logging

from pymilvus.bulk_writer.constants import ConnectType

logger = logging.getLogger("EndpointResolver")
logging.basicConfig(level=logging.INFO)


class EndpointResolver:
    @staticmethod
    def resolve_endpoint(
        default_endpoint: str, cloud: str, region: str, connect_type: ConnectType
    ) -> str:
        logger.info(
            "Start resolving endpoint, cloud=%s, region=%s, connectType=%s",
            cloud,
            region,
            connect_type,
        )
        if cloud == "ali":
            default_endpoint = EndpointResolver._resolve_oss_endpoint(region, connect_type)
        logger.info("Resolved endpoint: %s, reachable check passed", default_endpoint)
        return default_endpoint

    @staticmethod
    def _resolve_oss_endpoint(region: str, connect_type: ConnectType) -> str:
        internal_endpoint = f"oss-{region}-internal.aliyuncs.com"
        public_endpoint = f"oss-{region}.aliyuncs.com"

        if connect_type == ConnectType.INTERNAL:
            logger.info("Forced INTERNAL endpoint selected: %s", internal_endpoint)
            EndpointResolver._check_endpoint_reachable(internal_endpoint, True)
            return internal_endpoint
        if connect_type == ConnectType.PUBLIC:
            logger.info("Forced PUBLIC endpoint selected: %s", public_endpoint)
            EndpointResolver._check_endpoint_reachable(public_endpoint, True)
            return public_endpoint
        if EndpointResolver._check_endpoint_reachable(internal_endpoint, False):
            logger.info("AUTO mode: internal endpoint reachable, using %s", internal_endpoint)
            return internal_endpoint
        logger.warning(
            "AUTO mode: internal endpoint not reachable, fallback to public endpoint %s",
            public_endpoint,
        )
        EndpointResolver._check_endpoint_reachable(public_endpoint, True)
        return public_endpoint

    @staticmethod
    def _check_endpoint_reachable(endpoint: str, raise_error: bool) -> bool:
        try:
            conn = http.client.HTTPSConnection(endpoint, timeout=5)
            conn.request("HEAD", "/")
            resp = conn.getresponse()
            code = resp.status
            logger.debug("Checked endpoint %s, response code=%s", endpoint, code)
        except Exception as e:
            if raise_error:
                logger.exception("Endpoint %s not reachable, throwing exception", endpoint)
                raise RuntimeError(str(e)) from e
            logger.warning("Endpoint %s not reachable, will fallback if needed", endpoint)
            return False
        else:
            return 200 <= code < 400



from typing import Dict


class IndexParam:
    def __init__(self, field_name: str, index_type: str, index_name: str, **kwargs):
        """
        Examples:

            >>> IndexParam(
            >>>     field_name="embeddings",
            >>>     index_type="HNSW",
            >>>     index_name="hnsw_index",
            >>>     metric_type="COSINE",
            >>>     M=64,
            >>>     efConstruction=100,
            >>> )
        """
        self._field_name = field_name
        self._index_type = index_type
        self._index_name = index_name

        # index configs are unique to each index,
        # if params={} is passed in, it will be flattened and merged
        # with other configs.
        self._configs = {}
        self._configs.update(kwargs.pop("params", {}))
        self._configs.update(kwargs)

    @property
    def field_name(self):
        return self._field_name

    @property
    def index_name(self):
        return self._index_name

    @property
    def index_type(self):
        return self._index_type

    def get_index_configs(self) -> Dict:
        """return index_type and index configs in a dict

        Examples:

            {
                "index_type": "HNSW",
                "metrics_type": "COSINE",
                "M": 64,
                "efConstruction": 100,
            }
        """
        configs = self._configs
        # Omit index_type if it's empty
        if self.index_type:
            configs["index_type"] = self.index_type
        return configs

    def to_dict(self) -> Dict:
        """All params"""
        return {
            "field_name": self.field_name,
            "index_type": self.index_type,
            "index_name": self.index_name,
            **self._configs,
        }

    def __str__(self):
        return str(self.to_dict())

    __repr__ = __str__

    def __eq__(self, other: None):
        if isinstance(other, self.__class__):
            return dict(self) == dict(other)

        if isinstance(other, dict):
            return dict(self) == other
        return False


class IndexParams(list):
    """List of indexs of a collection"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def add_index(self, field_name: str, index_type: str = "", index_name: str = "", **kwargs):
        index_param = IndexParam(field_name, index_type, index_name, **kwargs)
        super().append(index_param)



import pathlib
import sys
from tempfile import TemporaryDirectory

import numpy as np
import pytest
from pymilvus.exceptions import ConnectionConfigException
from pymilvus.milvus_client import MilvusClient


@pytest.mark.skipif(
    sys.platform.startswith("win"), reason="Milvus Lite is not supported on Windows"
)
class TestMilvusLite:
    @pytest.mark.skip("Milvus Lite is now an optional dependency. This test will be fixed later.")
    def test_milvus_lite(self):
        with TemporaryDirectory(dir="./") as root:
            db_file = pathlib.Path(root).joinpath("test.db")
            client = MilvusClient(db_file.as_posix())
            client.create_collection(collection_name="demo_collection", dimension=3)

            # Text strings to search from.
            docs = [
                "Artificial intelligence was founded as an academic discipline in 1956.",
                "Alan Turing was the first person to conduct substantial research in AI.",
                "Born in Maida Vale, London, Turing was raised in southern England.",
            ]

            rng = np.random.default_rng(seed=19530)
            vectors = [[rng.uniform(-1, 1) for _ in range(3)] for _ in range(len(docs))]
            data = [
                {"id": i, "vector": vectors[i], "text": docs[i], "subject": "history"}
                for i in range(len(vectors))
            ]
            res = client.insert(collection_name="demo_collection", data=data)
            assert res["insert_count"] == 3

            res = client.search(
                collection_name="demo_collection",
                data=[vectors[0]],
                filter="subject == 'history'",
                limit=2,
                output_fields=["text", "subject"],
            )
            assert len(res[0]) == 2

            # a query that retrieves all entities matching filter expressions.
            res = client.query(
                collection_name="demo_collection",
                filter="subject == 'history'",
                output_fields=["text", "subject"],
            )
            assert len(res) == 3

            # delete
            res = client.delete(
                collection_name="demo_collection",
                filter="subject == 'history'",
            )
            assert len(res) == 3

    def test_illegal_name(self):
        with pytest.raises(ConnectionConfigException) as e:
            MilvusClient("localhost")
        # check the raised exception contained
        assert (
            e.value.message
            == "uri: localhost is illegal, needs start with [unix, http, https, tcp] or a local file endswith [.db]"
        )



import shutil
import tempfile
import threading
from pathlib import Path
from unittest.mock import MagicMock, Mock, PropertyMock, patch

import pytest
from pymilvus.bulk_writer.constants import MB, BulkFileType
from pymilvus.bulk_writer.local_bulk_writer import LocalBulkWriter
from pymilvus.client.types import DataType
from pymilvus.orm.schema import CollectionSchema, FieldSchema


class TestLocalBulkWriter:
    @pytest.fixture
    def simple_schema(self):
        fields = [
            FieldSchema(name="id", dtype=DataType.INT64, is_primary=True, auto_id=False),
            FieldSchema(name="vector", dtype=DataType.FLOAT_VECTOR, dim=128),
            FieldSchema(name="text", dtype=DataType.VARCHAR, max_length=512),
        ]
        return CollectionSchema(fields=fields)

    @pytest.fixture
    def temp_dir(self):
        """Create a temporary directory that gets cleaned up after test"""
        temp_path = tempfile.mkdtemp()
        yield temp_path
        # Cleanup after test
        if Path(temp_path).exists():
            shutil.rmtree(temp_path)

    @patch('uuid.uuid4')
    def test_init(self, mock_uuid, simple_schema, temp_dir):
        mock_uuid.return_value = "test-uuid"

        writer = LocalBulkWriter(
            schema=simple_schema,
            local_path=temp_dir,
            chunk_size=128 * MB,
            file_type=BulkFileType.JSON
        )

        # Check the UUID directory was created
        expected_path = Path(temp_dir) / "test-uuid"
        assert writer._local_path == expected_path
        assert expected_path.exists()
        assert writer._uuid == "test-uuid"
        assert writer._chunk_size == 128 * MB
        assert writer._file_type == BulkFileType.JSON
        assert writer._flush_count == 0

    def test_init_with_segment_size(self, simple_schema, temp_dir):
        """Test backward compatibility with segment_size parameter"""
        writer = LocalBulkWriter(
            schema=simple_schema,
            local_path=temp_dir,
            chunk_size=64 * MB,
            segment_size=256 * MB,  # Should override chunk_size
            file_type=BulkFileType.PARQUET
        )

        assert writer._chunk_size == 256 * MB

    def test_context_manager(self, simple_schema, temp_dir):
        uuid_dir = None
        with LocalBulkWriter(
            schema=simple_schema,
            local_path=temp_dir,
            chunk_size=128 * MB,
            file_type=BulkFileType.JSON
        ) as writer:
            assert writer is not None
            uuid_dir = writer._local_path
            assert uuid_dir.exists()

        # After context exit, empty UUID directory should be removed
        assert not uuid_dir.exists()

    def test_del_method(self, simple_schema, temp_dir):
        writer = LocalBulkWriter(
            schema=simple_schema,
            local_path=temp_dir,
            chunk_size=128 * MB,
            file_type=BulkFileType.JSON
        )
        uuid_dir = writer._local_path
        assert uuid_dir.exists()

        del writer
        # Empty UUID directory should be removed through __del__
        assert not uuid_dir.exists()

    @patch('pymilvus.bulk_writer.local_bulk_writer.Thread')
    def test_append_row_triggers_flush(self, mock_thread, simple_schema, temp_dir):
        mock_thread_instance = MagicMock()
        mock_thread.return_value = mock_thread_instance

        writer = LocalBulkWriter(
            schema=simple_schema,
            local_path=temp_dir,
            chunk_size=1,  # Very small size to trigger flush
            file_type=BulkFileType.JSON
        )

        # Use PropertyMock to mock the buffer_size property
        with patch.object(type(writer), 'buffer_size', new_callable=PropertyMock) as mock_buffer_size:
            mock_buffer_size.return_value = 2  # Larger than chunk_size
            writer.append_row({
                "id": 1,
                "vector": [1.0] * 128,
                "text": "test"
            })

        # Verify flush thread was started
        mock_thread.assert_called()
        mock_thread_instance.start.assert_called()

    @patch('pymilvus.bulk_writer.local_bulk_writer.Thread')
    def test_commit_sync(self, mock_thread, simple_schema, temp_dir):
        mock_thread_instance = MagicMock()
        mock_thread.return_value = mock_thread_instance

        writer = LocalBulkWriter(
            schema=simple_schema,
            local_path=temp_dir,
            chunk_size=128 * MB,
            file_type=BulkFileType.JSON
        )

        writer.append_row({
            "id": 1,
            "vector": [1.0] * 128,
            "text": "test"
        })

        # Commit synchronously
        writer.commit(_async=False)

        # Thread should be started and joined
        mock_thread.assert_called()
        mock_thread_instance.start.assert_called()
        mock_thread_instance.join.assert_called()

    @patch('time.sleep')
    @patch('pymilvus.bulk_writer.local_bulk_writer.Thread')
    def test_commit_async(self, mock_thread, mock_sleep, simple_schema, temp_dir):
        mock_thread_instance = MagicMock()
        mock_thread.return_value = mock_thread_instance

        writer = LocalBulkWriter(
            schema=simple_schema,
            local_path=temp_dir,
            chunk_size=128 * MB,
            file_type=BulkFileType.JSON
        )

        writer.append_row({
            "id": 1,
            "vector": [1.0] * 128,
            "text": "test"
        })

        # Commit asynchronously
        writer.commit(_async=True)

        # Thread should be started but not joined
        mock_thread.assert_called()
        mock_thread_instance.start.assert_called()
        mock_thread_instance.join.assert_not_called()

    def test_commit_with_callback(self, simple_schema, temp_dir):
        writer = LocalBulkWriter(
            schema=simple_schema,
            local_path=temp_dir,
            chunk_size=128 * MB,
            file_type=BulkFileType.JSON
        )

        callback_mock = Mock()

        # Mock the buffer to have data using PropertyMock
        with patch.object(type(writer._buffer), 'row_count', new_callable=PropertyMock) as mock_row_count:
            mock_row_count.return_value = 1
            with patch.object(writer._buffer, 'persist') as mock_persist:
                test_file = str(writer._local_path / "1" / "test.json")
                mock_persist.return_value = [test_file]
                # Add current thread to _working_thread to avoid KeyError
                writer._working_thread[threading.current_thread().name] = threading.current_thread()
                writer._flush(callback_mock)

                # Callback should be called with file list
                callback_mock.assert_called_once_with([test_file])

    def test_flush_with_empty_buffer(self, simple_schema, temp_dir):
        writer = LocalBulkWriter(
            schema=simple_schema,
            local_path=temp_dir,
            chunk_size=128 * MB,
            file_type=BulkFileType.JSON
        )

        # Mock empty buffer using PropertyMock
        with patch.object(type(writer._buffer), 'row_count', new_callable=PropertyMock) as mock_row_count:
            mock_row_count.return_value = 0
            # Add current thread to _working_thread to avoid KeyError
            writer._working_thread[threading.current_thread().name] = threading.current_thread()
            writer._flush()

        # No files should be added
        assert len(writer._local_files) == 0

    def test_flush_exception_handling(self, simple_schema, temp_dir):
        writer = LocalBulkWriter(
            schema=simple_schema,
            local_path=temp_dir,
            chunk_size=128 * MB,
            file_type=BulkFileType.JSON
        )

        # Mock buffer to raise exception using PropertyMock
        with patch.object(type(writer._buffer), 'row_count', new_callable=PropertyMock) as mock_row_count:
            mock_row_count.return_value = 1
            with patch.object(writer._buffer, 'persist', side_effect=Exception("Test error")):
                # Add current thread to _working_thread to avoid KeyError
                writer._working_thread[threading.current_thread().name] = threading.current_thread()
                with pytest.raises(Exception, match="Test error"):
                    writer._flush()

    def test_properties(self, simple_schema, temp_dir):
        writer = LocalBulkWriter(
            schema=simple_schema,
            local_path=temp_dir,
            chunk_size=128 * MB,
            file_type=BulkFileType.JSON
        )

        assert writer.uuid == writer._uuid
        assert writer.data_path == writer._local_path
        assert writer.batch_files == []

        # Add some files
        writer._local_files = [["file1.json"], ["file2.json"]]
        assert writer.batch_files == [["file1.json"], ["file2.json"]]

    def test_multiple_flushes(self, simple_schema, temp_dir):
        writer = LocalBulkWriter(
            schema=simple_schema,
            local_path=temp_dir,
            chunk_size=128 * MB,
            file_type=BulkFileType.JSON
        )

        # Simulate multiple flushes using PropertyMock
        with patch.object(type(writer._buffer), 'row_count', new_callable=PropertyMock) as mock_row_count:
            mock_row_count.return_value = 1
            with patch.object(writer._buffer, 'persist') as mock_persist:
                mock_persist.side_effect = [
                    [str(writer._local_path / "1" / "1.json")],
                    [str(writer._local_path / "2" / "2.json")],
                    [str(writer._local_path / "3" / "3.json")]
                ]

                # Add current thread to _working_thread before each flush
                current_thread = threading.current_thread()

                writer._working_thread[current_thread.name] = current_thread
                writer._flush()
                assert writer._flush_count == 1

                writer._working_thread[current_thread.name] = current_thread
                writer._flush()
                assert writer._flush_count == 2

                writer._working_thread[current_thread.name] = current_thread
                writer._flush()
                assert writer._flush_count == 3

                assert len(writer._local_files) == 3

    @patch('time.sleep')
    def test_wait_for_previous_flush(self, mock_sleep, simple_schema, temp_dir):
        writer = LocalBulkWriter(
            schema=simple_schema,
            local_path=temp_dir,
            chunk_size=128 * MB,
            file_type=BulkFileType.JSON
        )

        # Simulate a working thread
        mock_thread = MagicMock()
        writer._working_thread = {"thread1": mock_thread}

        # Create a side effect that clears the working thread after first call
        def sleep_side_effect(duration):
            if mock_sleep.call_count == 1:
                writer._working_thread.clear()

        mock_sleep.side_effect = sleep_side_effect

        writer.commit(_async=False)

        # Should have waited for previous thread
        mock_sleep.assert_called()

    def test_rm_dir(self, simple_schema, temp_dir):
        writer = LocalBulkWriter(
            schema=simple_schema,
            local_path=temp_dir,
            chunk_size=128 * MB,
            file_type=BulkFileType.JSON
        )

        uuid_dir = writer._local_path
        assert uuid_dir.exists()

        # Test removing empty directory
        writer._rm_dir()

        # Directory should be removed
        assert not uuid_dir.exists()

    def test_rm_dir_not_empty(self, simple_schema, temp_dir):
        writer = LocalBulkWriter(
            schema=simple_schema,
            local_path=temp_dir,
            chunk_size=128 * MB,
            file_type=BulkFileType.JSON
        )

        uuid_dir = writer._local_path

        # Create a file in the directory
        test_file = uuid_dir / "test.json"
        test_file.write_text("{}")

        assert uuid_dir.exists()
        assert test_file.exists()

        # Test removing non-empty directory
        writer._rm_dir()

        # Directory should NOT be removed (still has files)
        assert uuid_dir.exists()
        assert test_file.exists()

    def test_exit_waits_for_threads(self, simple_schema, temp_dir):
        writer = LocalBulkWriter(
            schema=simple_schema,
            local_path=temp_dir,
            chunk_size=128 * MB,
            file_type=BulkFileType.JSON
        )

        # Add mock working threads
        mock_thread1 = MagicMock()
        mock_thread2 = MagicMock()
        writer._working_thread = {
            "thread1": mock_thread1,
            "thread2": mock_thread2
        }

        writer._exit()

        # Should wait for all threads
        mock_thread1.join.assert_called_once()
        mock_thread2.join.assert_called_once()



import logging
import os
from unittest import mock

import pytest
from pymilvus import *
from pymilvus import DefaultConfig, MilvusException, connections
from pymilvus.exceptions import ErrorCode

LOGGER = logging.getLogger(__name__)

mock_prefix = "pymilvus.client.grpc_handler.GrpcHandler"


class TestConnect:
    """
        Connect to a connected alias will:
        - ignore and return if no configs are given
        - raise ConnectionConfigException if inconsistent configs are given

        Connect to an existing and not connected alias will:
        - connect with the existing alias config if no configs are given
        - connect with the providing configs if valid, and replace the old ones.

        Connect to a new alias will:
        - connect with the providing configs if valid, store the new alias with these configs

    """
    @pytest.fixture(scope="function", params=[
        {},
        {"random": "not useful"},
    ])
    def no_host_port(self, request):
        return request.param

    @pytest.fixture(scope="function", params=[
        {"port": "19530"},
        {"host": "localhost"},
    ])
    def no_host_or_port(self, request):
        return request.param

    @pytest.fixture(scope="function", params=[
        {"uri": "https://127.0.0.1:19530"},
        {"uri": "tcp://127.0.0.1:19530"},
        {"uri": "http://127.0.0.1:19530"},
        {"uri": "http://example.com:80"},
        {"uri": "http://example.com:80/database1"},
        {"uri": "https://127.0.0.1:19530/database2"},
        {"uri": "https://127.0.0.1/database3"},
        {"uri": "http://127.0.0.1/database4"},
    ])
    def uri(self, request):
        return request.param

    def test_connect_with_default_config(self):
        alias = "default"
        default_addr = {"address": "localhost:19530", "user": ""}

        assert connections.has_connection(alias) is False
        addr = connections.get_connection_addr(alias)

        assert addr == default_addr

        with mock.patch(f"{mock_prefix}.__init__", return_value=None), mock.patch(f"{mock_prefix}._wait_for_channel_ready", return_value=None):
                connections.connect(keep_alive=False)

        assert connections.has_connection(alias) is True

        addr = connections.get_connection_addr(alias)
        assert addr == default_addr

        with mock.patch(f"{mock_prefix}.close", return_value=None):
            connections.disconnect(alias)

    @pytest.fixture(scope="function", params=[
        ("", {"address": "localhost:19530", "user": ""}),
        ("localhost", {"address": "localhost:19530", "user": ""}),
        ("localhost:19530", {"address": "localhost:19530", "user": ""}),
        ("abc@localhost", {"address": "localhost:19530", "user": "abc"}),
        ("milvus_host", {"address": "milvus_host:19530", "user": ""}),
        ("milvus_host:12012", {"address": "milvus_host:12012", "user": ""}),
        ("abc@milvus_host:12012", {"address": "milvus_host:12012", "user": "abc"}),
        ("abc@milvus_host", {"address": "milvus_host:19530", "user": "abc"}),
    ])
    def test_connect_with_default_config_from_environment(self, env_result):
        os.environ[DefaultConfig.MILVUS_URI] = env_result[0]
        assert env_result[1] == connections._read_default_config_from_os_env()

        # use env
        with mock.patch(f"{mock_prefix}.__init__", return_value=None), mock.patch(f"{mock_prefix}._wait_for_channel_ready", return_value=None):
                connections.connect(keep_alive=False)

        assert env_result[1] == connections.get_connection_addr(DefaultConfig.MILVUS_CONN_ALIAS)

        # use param
        with mock.patch(f"{mock_prefix}.__init__", return_value=None), mock.patch(f"{mock_prefix}._wait_for_channel_ready", return_value=None):
                connections.connect(DefaultConfig.MILVUS_CONN_ALIAS, host="test_host", port="19999", keep_alive=False)

        curr_addr = connections.get_connection_addr(DefaultConfig.MILVUS_CONN_ALIAS)
        assert env_result[1] != curr_addr
        assert curr_addr == {"address":"test_host:19999", "user": ""}

        with mock.patch(f"{mock_prefix}.close", return_value=None):
            connections.remove_connection(DefaultConfig.MILVUS_CONN_ALIAS)

    def test_connect_new_alias_with_configs(self):
        alias = "exist"
        addr = {"address": "localhost:19530"}

        assert connections.has_connection(alias) is False
        a = connections.get_connection_addr(alias)
        assert a == {}

        with mock.patch(f"{mock_prefix}.__init__", return_value=None), mock.patch(f"{mock_prefix}._wait_for_channel_ready", return_value=None):
                connections.connect(alias, **addr, keep_alive=False)

        assert connections.has_connection(alias) is True

        a = connections.get_connection_addr(alias)
        a.pop("user")
        assert a == addr

        with mock.patch(f"{mock_prefix}.close", return_value=None):
            connections.remove_connection(alias)

    def test_connect_new_alias_with_configs_NoHostOrPort(self, no_host_or_port):
        alias = "no_host_or_port"

        assert connections.has_connection(alias) is False
        a = connections.get_connection_addr(alias)
        assert a == {}

        with mock.patch(f"{mock_prefix}.__init__", return_value=None), mock.patch(f"{mock_prefix}._wait_for_channel_ready", return_value=None):
                connections.connect(alias, **no_host_or_port, keep_alive=False)

        assert connections.has_connection(alias) is True
        assert connections.get_connection_addr(alias) == {"address": "localhost:19530", "user": ""}

        with mock.patch(f"{mock_prefix}.close", return_value=None):
            connections.remove_connection(alias)

    def test_connect_new_alias_with_no_config(self):
        alias = self.test_connect_new_alias_with_no_config.__name__

        assert connections.has_connection(alias) is False
        a = connections.get_connection_addr(alias)
        assert a == {}

        with pytest.raises(MilvusException) as excinfo:
            connections.connect(alias, keep_alive=False)

        LOGGER.info(f"Exception info: {excinfo.value}")
        assert "You need to pass in the configuration" in excinfo.value.message
        assert excinfo.value.code == ErrorCode.UNEXPECTED_ERROR

    def test_connect_with_uri(self, uri):
        alias = self.test_connect_with_uri.__name__

        with mock.patch(f"{mock_prefix}._setup_grpc_channel", return_value=None), mock.patch(f"{mock_prefix}._wait_for_channel_ready", return_value=None):
                connections.connect(alias, **uri, keep_alive=False)

        addr = connections.get_connection_addr(alias)
        LOGGER.debug(addr)

        assert connections.has_connection(alias) is True

        with mock.patch(f"{mock_prefix}.close", return_value=None):
            connections.remove_connection(alias)

    def test_add_connection_then_connect(self, uri):
        alias = self.test_add_connection_then_connect.__name__

        connections.add_connection(**{alias: uri})
        addr1 = connections.get_connection_addr(alias)
        LOGGER.debug(f"addr1: {addr1}")

        with mock.patch(f"{mock_prefix}._setup_grpc_channel", return_value=None), mock.patch(f"{mock_prefix}._wait_for_channel_ready", return_value=None):
                connections.connect(alias, keep_alive=False)

        addr2 = connections.get_connection_addr(alias)
        LOGGER.debug(f"addr2: {addr2}")

        assert addr1 == addr2

        with mock.patch(f"{mock_prefix}.close", return_value=None):
            connections.remove_connection(alias)


class TestAddConnection:
    @pytest.fixture(scope="function", params=[
        {"host": "localhost", "port": "19530"},
        {"host": "localhost", "port": "19531"},
        {"host": "localhost", "port": "19530", "random": "useless"},
    ])
    def host_port(self, request):
        return request.param

    @pytest.fixture(scope="function", params=[
        {"host": None, "port": "19530"},
        {"host": 1, "port": "19531"},
        {"host": 1.0, "port": "19530", "random": "useless"},
    ])
    def invalid_host(self, request):
        return request.param

    @pytest.fixture(scope="function", params=[
        {"host": "localhost", "port": None},
        {"host": "localhost", "port": 1.0},
        {"host": "localhost", "port": b'19530', "random": "useless"},
    ])
    def invalid_port(self, request):
        return request.param

    def test_add_connection_no_error(self, host_port):
        add_connection = connections.add_connection

        add_connection(test=host_port)
        assert connections.get_connection_addr("test").get("address") == f"{host_port['host']}:{host_port['port']}"

        connections.remove_connection("test")

    def test_add_connection_no_error_with_user(self):
        add_connection = connections.add_connection

        host_port = {"host": "localhost", "port": "19530", "user": "_user"}

        add_connection(test=host_port)

        config = connections.get_connection_addr("test")
        assert config.get("address") == f"{host_port['host']}:{host_port['port']}"
        assert config.get("user") == host_port['user']

        add_connection(default=host_port)
        config = connections.get_connection_addr("default")
        assert config.get("address") == f"{host_port['host']}:{host_port['port']}"
        assert config.get("user") == host_port['user']

        connections.remove_connection("test")
        connections.disconnect("default")

    def test_add_connection_raise_HostType(self, invalid_host):
        add_connection = connections.add_connection

        with pytest.raises(MilvusException) as excinfo:
            add_connection(test=invalid_host)

        LOGGER.info(f"Exception info: {excinfo.value}")
        assert "Type of 'host' must be str." in excinfo.value.message
        assert excinfo.value.code == ErrorCode.UNEXPECTED_ERROR

    def test_add_connection_raise_PortType(self, invalid_port):
        add_connection = connections.add_connection

        with pytest.raises(MilvusException) as excinfo:
            add_connection(test=invalid_port)

        LOGGER.info(f"Exception info: {excinfo.value}")
        assert "Type of 'port' must be str" in excinfo.value.message
        assert excinfo.value.code == ErrorCode.UNEXPECTED_ERROR

    @pytest.mark.parametrize("valid_addr", [
        {"address": "127.0.0.1:19530"},
        {"address": "example.com:19530"},
    ])
    def test_add_connection_address(self, valid_addr):
        alias = self.test_add_connection_address.__name__
        config = {alias: valid_addr}
        connections.add_connection(**config)

        addr = connections.get_connection_addr(alias)
        assert addr.get("address") == valid_addr.get("address")
        LOGGER.info(f"addr: {addr}")

        with mock.patch(f"{mock_prefix}.close", return_value=None):
            connections.remove_connection(alias)

    @pytest.mark.parametrize("invalid_addr", [
        {"address": "127.0.0.1"},
        {"address": "19530"},
    ])
    def test_add_connection_address_invalid(self, invalid_addr):
        alias = self.test_add_connection_address_invalid.__name__
        config = {alias: invalid_addr}

        with pytest.raises(MilvusException) as excinfo:
            connections.add_connection(**config)

        LOGGER.info(f"Exception info: {excinfo.value}")
        assert "Illegal address" in excinfo.value.message
        assert excinfo.value.code == ErrorCode.UNEXPECTED_ERROR

    @pytest.mark.parametrize("valid_uri", [
        {"uri": "http://127.0.0.1:19530"},
        {"uri": "http://localhost:19530"},
        {"uri": "http://example.com:80"},
        {"uri": "http://example.com"},
    ])
    def test_add_connection_uri(self, valid_uri):
        alias = self.test_add_connection_uri.__name__
        config = {alias: valid_uri}
        connections.add_connection(**config)

        addr = connections.get_connection_addr(alias)
        LOGGER.info(f"addr: {addr}")

        host, port = addr["address"].split(':')
        assert host in valid_uri['uri'] or host in DefaultConfig.DEFAULT_HOST
        assert port in valid_uri['uri'] or port in DefaultConfig.DEFAULT_PORT
        LOGGER.info(f"host: {host}, port: {port}")

        with mock.patch(f"{mock_prefix}.close", return_value=None):
            connections.remove_connection(alias)

    @pytest.mark.parametrize("invalid_uri", [
        {"uri": "http://"},
        {"uri": None},
        {"uri": -1},
    ])
    def test_add_connection_uri_invalid(self, invalid_uri):
        alias = self.test_add_connection_uri_invalid.__name__
        config = {alias: invalid_uri}

        with pytest.raises(MilvusException) as excinfo:
            connections.add_connection(**config)

        LOGGER.info(f"Exception info: {excinfo.value}")
        assert "Illegal uri" in excinfo.value.message
        assert excinfo.value.code == ErrorCode.UNEXPECTED_ERROR


class TestIssues:
    def test_issue_1196(self):
        """
        >>> connections.connect(alias="default11", host="xxx.com", port=19541, user="root", password="12345", secure=True)
        >>> connections.add_connection(default={"host": "xxx.com", "port": 19541})
        >>> connections.connect("default", user="root", password="12345", secure=True)
        Traceback (most recent call last):
          File "/usr/local/lib/python3.8/dist-packages/pymilvus/client/grpc_handler.py", line 114, in _wait_for_channel_ready
            grpc.channel_ready_future(self._channel).result(timeout=3)
          File "/usr/local/lib/python3.8/dist-packages/grpc/_utilities.py", line 139, in result
            self._block(timeout)
          File "/usr/local/lib/python3.8/dist-packages/grpc/_utilities.py", line 85, in _block
            raise grpc.FutureTimeoutError()
        grpc.FutureTimeoutError
        """

        alias = self.test_issue_1196.__name__

        with mock.patch(f"{mock_prefix}.__init__", return_value=None), mock.patch(f"{mock_prefix}._wait_for_channel_ready", return_value=None):
                config = {"alias": alias, "host": "localhost", "port": "19531", "user": "root", "password": 12345, "secure": True}
                connections.connect(**config, keep_alive=False)
                config = connections.get_connection_addr(alias)
                assert config == {"address": 'localhost:19531', "user": 'root', "secure": True}

                connections.add_connection(default={"host": "localhost", "port": 19531})
                config = connections.get_connection_addr("default")
                assert config == {"address": 'localhost:19531', "user": ""}

                connections.connect("default", user="root", password="12345", secure=True, keep_alive=False)

                config = connections.get_connection_addr("default")
                assert config == {"address": 'localhost:19531', "user": 'root', "secure": True}

    @pytest.mark.parametrize("uri, db_name, expected_db_name", [
        # Issue #2670: URI ending with slash should not overwrite explicit db_name
        ("http://localhost:19530/", "test_db", "test_db"),
        ("https://localhost:19530/", "production_db", "production_db"),
        ("tcp://localhost:19530/", "test_db", "test_db"),

        # Issue #2727: db_name passed in URI path should be used when no explicit db_name
        ("http://localhost:19530/test_db", "", "test_db"),
        ("http://localhost:19530/production_db", "", "production_db"),
        ("https://localhost:19530/test_db", "", "test_db"),

        # Mixed scenarios: explicit db_name takes precedence over URI path
        ("http://localhost:19530/uri_db", "explicit_db", "explicit_db"),
        ("http://localhost:19530", "test_db", "test_db"),
        ("http://localhost:19530", "", "default"),

        # Multiple path segments - only first should be used as db_name
        ("http://localhost:19530/db1/collection1", "", "db1"),
        ("http://localhost:19530/db1/collection1/", "", "db1"),

        # Empty path segments should be handled correctly
        ("http://localhost:19530//", "test_db", "test_db"),
        ("http://localhost:19530///", "test_db", "test_db"),
    ])
    def test_issue_2670_2727(self, uri: str, db_name: str, expected_db_name: str):
        """
        Issue 2670:
        Test for db_name being overwritten with empty string, when the uri
        ends in a slash - e.g. http://localhost:19530/

        See: https://github.com/milvus-io/pymilvus/issues/2670

        Actual behaviour before fix: if a uri is passed ending with a slash,
            it will overwrite the db_name with an empty string.
        Expected and current behaviour: if db_name is passed explicitly,
            it should be used in the initialization of the GrpcHandler.

        Issue 2727:
        If db_name is passed as a path to the uri and not explicitly passed as an argument,
        it is not overwritten with an empty string.

        See: https://github.com/milvus-io/pymilvus/issues/2727

        Actual behaviour before fix: if db_name is passed as a path to the uri,
            it will overwrite the db_name with an empty string.
        Expected and current behaviour: if db_name is passed as a path to the uri,
            it should be used in the initialization of the GrpcHandler.
        """
        alias = f"test_2670_2727_{uri.replace('://', '_').replace('/', '_')}_{db_name}"

        with mock.patch(f"{mock_prefix}.__init__", return_value=None) as mock_init, mock.patch(
            f"{mock_prefix}._wait_for_channel_ready", return_value=None):
            config = {"alias": alias, "uri": uri}
            # Always pass db_name parameter, even if it's an empty string
            if db_name or db_name == "":  # Pass both empty and non-empty strings
                config["db_name"] = db_name
            connections.connect(**config, keep_alive=False)

            # Verify that GrpcHandler was initialized with the correct db_name
            mock_init.assert_called_once()
            call_args = mock_init.call_args
            actual_db_name = call_args.kwargs.get("db_name", "default")

            assert actual_db_name == expected_db_name, (
                f"Expected db_name to be '{expected_db_name}', "
                f"but got '{actual_db_name}' for uri='{uri}' and db_name='{db_name}'"
            )

            # Clean up - mock the close method to avoid AttributeError
            with mock.patch(f"{mock_prefix}.close", return_value=None):
                connections.remove_connection(alias)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: feder.proto
# Protobuf Python Version: 5.27.2
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    27,
    2,
    '',
    'feder.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import common_pb2 as common__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0b\x66\x65\x64\x65r.proto\x12\x12milvus.proto.feder\x1a\x0c\x63ommon.proto\"9\n\x10SegmentIndexData\x12\x11\n\tsegmentID\x18\x01 \x01(\x03\x12\x12\n\nindex_data\x18\x02 \x01(\t\"A\n\x18\x46\x65\x64\x65rSegmentSearchResult\x12\x11\n\tsegmentID\x18\x01 \x01(\x03\x12\x12\n\nvisit_info\x18\x02 \x01(\t\"t\n\x19ListIndexedSegmentRequest\x12*\n\x04\x62\x61se\x18\x01 \x01(\x0b\x32\x1c.milvus.proto.common.MsgBase\x12\x17\n\x0f\x63ollection_name\x18\x02 \x01(\t\x12\x12\n\nindex_name\x18\x03 \x01(\t\"]\n\x1aListIndexedSegmentResponse\x12+\n\x06status\x18\x01 \x01(\x0b\x32\x1b.milvus.proto.common.Status\x12\x12\n\nsegmentIDs\x18\x02 \x03(\x03\"\x8f\x01\n\x1f\x44\x65scribeSegmentIndexDataRequest\x12*\n\x04\x62\x61se\x18\x01 \x01(\x0b\x32\x1c.milvus.proto.common.MsgBase\x12\x17\n\x0f\x63ollection_name\x18\x02 \x01(\t\x12\x12\n\nindex_name\x18\x03 \x01(\t\x12\x13\n\x0bsegmentsIDs\x18\x04 \x03(\x03\"\xb9\x02\n DescribeSegmentIndexDataResponse\x12+\n\x06status\x18\x01 \x01(\x0b\x32\x1b.milvus.proto.common.Status\x12W\n\nindex_data\x18\x02 \x03(\x0b\x32\x43.milvus.proto.feder.DescribeSegmentIndexDataResponse.IndexDataEntry\x12\x37\n\x0cindex_params\x18\x03 \x03(\x0b\x32!.milvus.proto.common.KeyValuePair\x1aV\n\x0eIndexDataEntry\x12\x0b\n\x03key\x18\x01 \x01(\x03\x12\x33\n\x05value\x18\x02 \x01(\x0b\x32$.milvus.proto.feder.SegmentIndexData:\x02\x38\x01\x42\x35Z3github.com/milvus-io/milvus-proto/go-api/v2/federpbb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'feder_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'Z3github.com/milvus-io/milvus-proto/go-api/v2/federpb'
  _globals['_DESCRIBESEGMENTINDEXDATARESPONSE_INDEXDATAENTRY']._loaded_options = None
  _globals['_DESCRIBESEGMENTINDEXDATARESPONSE_INDEXDATAENTRY']._serialized_options = b'8\001'
  _globals['_SEGMENTINDEXDATA']._serialized_start=49
  _globals['_SEGMENTINDEXDATA']._serialized_end=106
  _globals['_FEDERSEGMENTSEARCHRESULT']._serialized_start=108
  _globals['_FEDERSEGMENTSEARCHRESULT']._serialized_end=173
  _globals['_LISTINDEXEDSEGMENTREQUEST']._serialized_start=175
  _globals['_LISTINDEXEDSEGMENTREQUEST']._serialized_end=291
  _globals['_LISTINDEXEDSEGMENTRESPONSE']._serialized_start=293
  _globals['_LISTINDEXEDSEGMENTRESPONSE']._serialized_end=386
  _globals['_DESCRIBESEGMENTINDEXDATAREQUEST']._serialized_start=389
  _globals['_DESCRIBESEGMENTINDEXDATAREQUEST']._serialized_end=532
  _globals['_DESCRIBESEGMENTINDEXDATARESPONSE']._serialized_start=535
  _globals['_DESCRIBESEGMENTINDEXDATARESPONSE']._serialized_end=848
  _globals['_DESCRIBESEGMENTINDEXDATARESPONSE_INDEXDATAENTRY']._serialized_start=762
  _globals['_DESCRIBESEGMENTINDEXDATARESPONSE_INDEXDATAENTRY']._serialized_end=848
# @@protoc_insertion_point(module_scope)



# Copyright (C) 2019-2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

import copy
from typing import Dict, Optional, TypeVar

from pymilvus.exceptions import CollectionNotExistException, ExceptionsMessage
from pymilvus.settings import Config

Index = TypeVar("Index")
Collection = TypeVar("Collection")


class Index:
    def __init__(
        self,
        collection: Collection,
        field_name: str,
        index_params: Dict,
        **kwargs,
    ) -> Index:
        """Creates index on a specified field according to the index parameters.

        Args:
            collection(Collection): The collection in which the index is created
            field_name(str): The name of the field to create an index for.
            index_params(dict): Indexing parameters.
            kwargs:
                * *index_name* (``str``) --
                  The name of index which will be created. If no index name is specified,
                  default index name will be used.

        Raises:
            MilvusException: If anything goes wrong.

        Examples:
            >>> from pymilvus import *
            >>> from pymilvus.schema import *
            >>> from pymilvus.types import DataType
            >>> connections.connect()
            <pymilvus.client.stub.Milvus object at 0x7fac15e53470>
            >>> field1 = FieldSchema("int64", DataType.INT64, is_primary=True)
            >>> field2 = FieldSchema("fvec", DataType.FLOAT_VECTOR, is_primary=False, dim=128)
            >>> schema = CollectionSchema(fields=[field1, field2])
            >>> collection = Collection(name='test_collection', schema=schema)
            >>> # insert some data
            >>> index_params = {
            ...     "index_type": "IVF_FLAT",
            ...     "metric_type": "L2",
            ...     "params": {"nlist": 128}}
            >>> index = Index(collection, "fvec", index_params)
            >>> index.params
            {'index_type': 'IVF_FLAT', 'metric_type': 'L2', 'params': {'nlist': 128}}
            >>> index.collection_name
            test_collection
            >>> index.field_name
            fvec
            >>> index.drop()
        """
        # ruff: noqa: PLC0415
        from .collection import Collection

        if not isinstance(collection, Collection):
            raise CollectionNotExistException(message=ExceptionsMessage.CollectionType)
        self._collection = collection
        self._field_name = field_name
        self._index_params = index_params
        self._index_name = kwargs.get("index_name", Config.IndexName)
        if kwargs.get("construct_only", False):
            return

        conn = self._get_connection()
        conn.create_index(self._collection.name, self._field_name, self._index_params, **kwargs)
        indexes = conn.list_indexes(self._collection.name)
        for index in indexes:
            if index.field_name == self._field_name:
                self._index_name = index.index_name
                break

    def _get_connection(self):
        return self._collection._get_connection()

    @property
    def params(self) -> dict:
        """dict: The index parameters"""
        return copy.deepcopy(self._index_params)

    @property
    def collection_name(self) -> str:
        """str: The corresponding collection name"""
        return self._collection.name

    @property
    def field_name(self) -> str:
        """str: The corresponding field name."""
        return self._field_name

    @property
    def index_name(self) -> str:
        """str: The corresponding index name."""
        return self._index_name

    def __eq__(self, other: Index) -> bool:
        """The order of the fields of index must be consistent."""
        return self.to_dict() == other.to_dict()

    def to_dict(self):
        """Put collection name, field name and index params into dict."""
        return {
            "collection": self._collection._name,
            "field": self._field_name,
            "index_name": self._index_name,
            "index_param": self.params,
        }

    def drop(self, timeout: Optional[float] = None, **kwargs):
        """Drop an index and its corresponding index files.

        Args:
            timeout(float, optional): An optional duration of time in seconds to allow
                for the RPC. When timeout is set to None, client waits until server response
                or error occur
        """
        conn = self._get_connection()
        conn.drop_index(
            collection_name=self._collection.name,
            field_name=self.field_name,
            index_name=self.index_name,
            timeout=timeout,
        )



import tempfile
from pathlib import Path
from typing import Any, Dict
from unittest.mock import MagicMock, Mock, patch

import pytest
import requests
from pymilvus.bulk_writer.constants import BulkFileType, ConnectType
from pymilvus.bulk_writer.stage_bulk_writer import StageBulkWriter
from pymilvus.bulk_writer.stage_file_manager import StageFileManager
from pymilvus.bulk_writer.stage_manager import StageManager
from pymilvus.bulk_writer.stage_restful import (
    apply_stage,
    create_stage,
    delete_stage,
    list_stages,
)
from pymilvus.client.types import DataType
from pymilvus.exceptions import MilvusException
from pymilvus.orm.schema import CollectionSchema, FieldSchema


class TestStageRestful:
    """Test stage RESTful API functions."""

    @pytest.fixture
    def mock_response(self) -> Mock:
        """Create a mock response object."""
        response = Mock(spec=requests.Response)
        response.status_code = 200
        response.json.return_value = {"code": 0, "message": "success", "data": {}}
        return response

    @pytest.fixture
    def api_params(self) -> Dict[str, str]:
        """Common API parameters."""
        return {
            "url": "https://api.cloud.zilliz.com",
            "api_key": "test_api_key",
        }

    @patch("pymilvus.bulk_writer.stage_restful.requests.get")
    def test_list_stages_success(
        self, mock_get: Mock, mock_response: Mock, api_params: Dict[str, str]
    ) -> None:
        """Test successful list_stages call."""
        mock_get.return_value = mock_response
        mock_response.json.return_value = {
            "code": 0,
            "message": "success",
            "data": {"stages": ["stage1", "stage2"]},
        }

        response = list_stages(
            **api_params,
            project_id="test_project",
            current_page=1,
            page_size=10,
        )

        assert response.status_code == 200
        assert response.json()["data"]["stages"] == ["stage1", "stage2"]
        mock_get.assert_called_once()

    @patch("pymilvus.bulk_writer.stage_restful.requests.get")
    def test_list_stages_failure(
        self, mock_get: Mock, mock_response: Mock, api_params: Dict[str, str]
    ) -> None:
        """Test failed list_stages call."""
        mock_response.json.return_value = {
            "code": 1001,
            "message": "Invalid API key",
            "data": {},
        }
        mock_get.return_value = mock_response

        with pytest.raises(MilvusException, match="Invalid API key"):
            list_stages(**api_params, project_id="test_project")

    @patch("pymilvus.bulk_writer.stage_restful.requests.post")
    def test_create_stage_success(
        self, mock_post: Mock, mock_response: Mock, api_params: Dict[str, str]
    ) -> None:
        """Test successful create_stage call."""
        mock_post.return_value = mock_response
        mock_response.json.return_value = {
            "code": 0,
            "message": "success",
            "data": {"stageId": "stage123"},
        }

        response = create_stage(
            **api_params,
            project_id="test_project",
            region_id="us-west-2",
            stage_name="test_stage",
        )

        assert response.status_code == 200
        assert response.json()["data"]["stageId"] == "stage123"
        mock_post.assert_called_once()

    @patch("pymilvus.bulk_writer.stage_restful.requests.delete")
    def test_delete_stage_success(
        self, mock_delete: Mock, mock_response: Mock, api_params: Dict[str, str]
    ) -> None:
        """Test successful delete_stage call."""
        mock_delete.return_value = mock_response

        response = delete_stage(**api_params, stage_name="test_stage")

        assert response.status_code == 200
        mock_delete.assert_called_once()

    @patch("pymilvus.bulk_writer.stage_restful.requests.post")
    def test_apply_stage_success(
        self, mock_post: Mock, mock_response: Mock, api_params: Dict[str, str]
    ) -> None:
        """Test successful apply_stage call."""
        mock_post.return_value = mock_response
        mock_response.json.return_value = {
            "code": 0,
            "message": "success",
            "data": {
                "stageName": "test_stage",
                "stagePrefix": "prefix/",
                "endpoint": "s3.amazonaws.com",
                "bucketName": "test-bucket",
                "region": "us-west-2",
                "cloud": "aws",
                "condition": {"maxContentLength": 1073741824},
                "credentials": {
                    "tmpAK": "test_access_key",
                    "tmpSK": "test_secret_key",
                    "sessionToken": "test_token",
                    "expireTime": "2024-12-31T23:59:59Z",
                },
            },
        }

        response = apply_stage(
            **api_params,
            stage_name="test_stage",
            path="data/",
        )

        assert response.status_code == 200
        data = response.json()["data"]
        assert data["stageName"] == "test_stage"
        assert data["endpoint"] == "s3.amazonaws.com"
        mock_post.assert_called_once()

    @patch("pymilvus.bulk_writer.stage_restful.requests.get")
    def test_http_error_handling(
        self, mock_get: Mock, api_params: Dict[str, str]
    ) -> None:
        """Test HTTP error handling."""
        mock_get.return_value.status_code = 404

        with pytest.raises(MilvusException, match="status code: 404"):
            list_stages(**api_params, project_id="test_project")

    @patch("pymilvus.bulk_writer.stage_restful.requests.get")
    def test_network_error_handling(
        self, mock_get: Mock, api_params: Dict[str, str]
    ) -> None:
        """Test network error handling."""
        mock_get.side_effect = requests.exceptions.ConnectionError("Network error")

        with pytest.raises(MilvusException, match="Network error"):
            list_stages(**api_params, project_id="test_project")


class TestStageManager:
    """Test StageManager class."""

    @pytest.fixture
    def stage_manager(self) -> StageManager:
        """Create a StageManager instance."""
        return StageManager(
            cloud_endpoint="https://api.cloud.zilliz.com",
            api_key="test_api_key",
        )

    @patch("pymilvus.bulk_writer.stage_manager.create_stage")
    def test_create_stage(self, mock_create: Mock, stage_manager: StageManager) -> None:
        """Test creating a stage."""
        stage_manager.create_stage(
            project_id="test_project",
            region_id="us-west-2",
            stage_name="test_stage",
        )

        mock_create.assert_called_once_with(
            stage_manager.cloud_endpoint,
            stage_manager.api_key,
            "test_project",
            "us-west-2",
            "test_stage",
        )

    @patch("pymilvus.bulk_writer.stage_manager.delete_stage")
    def test_delete_stage(self, mock_delete: Mock, stage_manager: StageManager) -> None:
        """Test deleting a stage."""
        stage_manager.delete_stage(stage_name="test_stage")

        mock_delete.assert_called_once_with(
            stage_manager.cloud_endpoint,
            stage_manager.api_key,
            "test_stage",
        )

    @patch("pymilvus.bulk_writer.stage_manager.list_stages")
    def test_list_stages(self, mock_list: Mock, stage_manager: StageManager) -> None:
        """Test listing stages."""
        mock_response = Mock()
        mock_response.json.return_value = {"data": {"stages": ["stage1", "stage2"]}}
        mock_list.return_value = mock_response

        result = stage_manager.list_stages(project_id="test_project", current_page=1, page_size=10)

        assert result.json()["data"]["stages"] == ["stage1", "stage2"]
        mock_list.assert_called_once_with(
            stage_manager.cloud_endpoint,
            stage_manager.api_key,
            "test_project",
            1,
            10,
        )


class TestStageFileManager:
    """Test StageFileManager class."""

    @pytest.fixture
    def stage_file_manager(self) -> StageFileManager:
        """Create a StageFileManager instance."""
        return StageFileManager(
            cloud_endpoint="https://api.cloud.zilliz.com",
            api_key="test_api_key",
            stage_name="test_stage",
            connect_type=ConnectType.AUTO,
        )

    @pytest.fixture
    def mock_stage_info(self) -> Dict[str, Any]:
        """Mock stage information."""
        return {
            "stageName": "test_stage",
            "stagePrefix": "prefix/",
            "endpoint": "s3.amazonaws.com",
            "bucketName": "test-bucket",
            "region": "us-west-2",
            "cloud": "aws",
            "condition": {"maxContentLength": 1073741824},
            "credentials": {
                "tmpAK": "test_access_key",
                "tmpSK": "test_secret_key",
                "sessionToken": "test_token",
                "expireTime": "2099-12-31T23:59:59Z",
            },
        }

    def test_convert_dir_path(self, stage_file_manager: StageFileManager) -> None:
        """Test directory path conversion."""
        assert stage_file_manager._convert_dir_path("") == ""
        assert stage_file_manager._convert_dir_path("/") == ""
        assert stage_file_manager._convert_dir_path("data") == "data/"
        assert stage_file_manager._convert_dir_path("data/") == "data/"

    @patch("pymilvus.bulk_writer.stage_file_manager.apply_stage")
    @patch("pymilvus.bulk_writer.stage_file_manager.Minio")
    def test_refresh_stage_and_client(
        self,
        mock_minio: Mock,
        mock_apply: Mock,
        stage_file_manager: StageFileManager,
        mock_stage_info: Dict[str, Any],
    ) -> None:
        """Test refreshing stage info and client."""
        mock_response = Mock()
        mock_response.json.return_value = {"data": mock_stage_info}
        mock_apply.return_value = mock_response

        stage_file_manager._refresh_stage_and_client("data/")

        assert stage_file_manager.stage_info == mock_stage_info
        mock_apply.assert_called_once()
        mock_minio.assert_called_once()

    def test_validate_size_success(
        self, stage_file_manager: StageFileManager, mock_stage_info: Dict[str, Any]
    ) -> None:
        """Test successful size validation."""
        stage_file_manager.stage_info = mock_stage_info
        stage_file_manager.total_bytes = 1000000  # 1MB

        # Should not raise any exception
        stage_file_manager._validate_size()

    def test_validate_size_failure(
        self, stage_file_manager: StageFileManager, mock_stage_info: Dict[str, Any]
    ) -> None:
        """Test size validation failure."""
        stage_file_manager.stage_info = mock_stage_info
        stage_file_manager.total_bytes = 2147483648  # 2GB

        with pytest.raises(ValueError, match="exceeds the maximum contentLength limit"):
            stage_file_manager._validate_size()

    @patch("pymilvus.bulk_writer.stage_file_manager.FileUtils.process_local_path")
    @patch.object(StageFileManager, "_refresh_stage_and_client")
    @patch.object(StageFileManager, "_validate_size")
    @patch.object(StageFileManager, "_put_object")
    def test_upload_file_to_stage(
        self,
        mock_put_object: Mock,
        mock_validate: Mock,
        mock_refresh: Mock,
        mock_process: Mock,
        stage_file_manager: StageFileManager,
        mock_stage_info: Dict[str, Any],
    ) -> None:
        """Test uploading file to stage."""
        with tempfile.TemporaryDirectory() as temp_dir:
            # Create test files
            test_file = Path(temp_dir) / "test.txt"
            test_file.write_text("test content")

            mock_process.return_value = ([str(test_file)], 12)
            stage_file_manager.stage_info = mock_stage_info

            result = stage_file_manager.upload_file_to_stage(str(test_file), "data/")

            assert result["stageName"] == "test_stage"
            assert result["path"] == "data/"
            mock_refresh.assert_called_once_with("data/")
            mock_validate.assert_called_once()

    @patch.object(StageFileManager, "_upload_with_retry")
    @patch.object(StageFileManager, "_refresh_stage_and_client")
    def test_put_object_refresh_on_expiry(
        self,
        mock_refresh: Mock,
        mock_upload: Mock,
        stage_file_manager: StageFileManager,
        mock_stage_info: Dict[str, Any],
    ) -> None:
        """Test that credentials are refreshed when expired."""
        # Set expired credentials
        expired_info = mock_stage_info.copy()
        expired_info["credentials"]["expireTime"] = "2020-01-01T00:00:00Z"
        stage_file_manager.stage_info = expired_info

        stage_file_manager._put_object("test.txt", "remote/test.txt", "data/")

        mock_refresh.assert_called_once_with("data/")
        mock_upload.assert_called_once()

    @patch("pymilvus.bulk_writer.stage_file_manager.Minio")
    def test_upload_with_retry_success(
        self, mock_minio: Mock, stage_file_manager: StageFileManager, mock_stage_info: Dict[str, Any]
    ) -> None:
        """Test successful upload with retry."""
        stage_file_manager.stage_info = mock_stage_info
        stage_file_manager._client = mock_minio.return_value

        stage_file_manager._upload_with_retry("test.txt", "remote/test.txt", "data/")

        stage_file_manager._client.fput_object.assert_called_once_with(
            bucket_name="test-bucket",
            object_name="remote/test.txt",
            file_path="test.txt",
        )

    @patch("pymilvus.bulk_writer.stage_file_manager.Minio")
    @patch.object(StageFileManager, "_refresh_stage_and_client")
    def test_upload_with_retry_failure(
        self,
        mock_refresh: Mock,
        mock_minio: Mock,
        stage_file_manager: StageFileManager,
        mock_stage_info: Dict[str, Any],
    ) -> None:
        """Test upload failure after max retries."""
        stage_file_manager.stage_info = mock_stage_info
        mock_client = mock_minio.return_value
        mock_client.fput_object.side_effect = Exception("Upload failed")
        stage_file_manager._client = mock_client

        with pytest.raises(RuntimeError, match="Upload failed after 2 attempts"):
            stage_file_manager._upload_with_retry("test.txt", "remote/test.txt", "data/", max_retries=2)

        assert mock_client.fput_object.call_count == 2
        assert mock_refresh.call_count == 2  # Refreshed on each retry


class TestStageBulkWriter:
    """Test StageBulkWriter class."""

    @pytest.fixture
    def simple_schema(self) -> CollectionSchema:
        """Create a simple collection schema."""
        fields = [
            FieldSchema(name="id", dtype=DataType.INT64, is_primary=True, auto_id=False),
            FieldSchema(name="vector", dtype=DataType.FLOAT_VECTOR, dim=128),
            FieldSchema(name="text", dtype=DataType.VARCHAR, max_length=512),
        ]
        return CollectionSchema(fields=fields)

    @pytest.fixture
    def stage_bulk_writer(self, simple_schema: CollectionSchema) -> StageBulkWriter:
        """Create a StageBulkWriter instance."""
        with patch("pymilvus.bulk_writer.stage_bulk_writer.StageFileManager"):
            return StageBulkWriter(
                schema=simple_schema,
                remote_path="test/data",
                cloud_endpoint="https://api.cloud.zilliz.com",
                api_key="test_api_key",
                stage_name="test_stage",
                chunk_size=1024,
                file_type=BulkFileType.PARQUET,
            )

    def test_init(self, stage_bulk_writer: StageBulkWriter) -> None:
        """Test StageBulkWriter initialization."""
        assert stage_bulk_writer._remote_path.endswith("/")
        assert stage_bulk_writer._stage_name == "test_stage"
        assert isinstance(stage_bulk_writer._stage_file_manager, MagicMock)

    def test_append_row(self, stage_bulk_writer: StageBulkWriter) -> None:
        """Test appending a row."""
        row = {
            "id": 1,
            "vector": [1.0] * 128,
            "text": "test text",
        }
        stage_bulk_writer.append_row(row)
        assert stage_bulk_writer.total_row_count == 1

    @patch.object(StageBulkWriter, "_upload")
    def test_commit(self, mock_upload: Mock, stage_bulk_writer: StageBulkWriter) -> None:
        """Test committing data."""
        # Add some data
        for i in range(10):
            stage_bulk_writer.append_row({
                "id": i,
                "vector": [float(i)] * 128,
                "text": f"text_{i}",
            })

        # Mock the upload to return file paths
        mock_upload.return_value = ["file1.parquet", "file2.parquet"]

        # Commit the data
        stage_bulk_writer.commit()

        # Upload should have been called during commit
        assert mock_upload.called

    def test_data_path_property(self, stage_bulk_writer: StageBulkWriter) -> None:
        """Test data_path property."""
        assert isinstance(stage_bulk_writer.data_path, str)
        assert "/" in stage_bulk_writer.data_path

    def test_batch_files_property(self, stage_bulk_writer: StageBulkWriter) -> None:
        """Test batch_files property."""
        assert stage_bulk_writer.batch_files == []
        stage_bulk_writer._remote_files = [["file1.parquet"], ["file2.parquet"]]
        assert stage_bulk_writer.batch_files == [["file1.parquet"], ["file2.parquet"]]

    def test_get_stage_upload_result(self, stage_bulk_writer: StageBulkWriter) -> None:
        """Test getting stage upload result."""
        result = stage_bulk_writer.get_stage_upload_result()
        assert result["stage_name"] == "test_stage"
        assert "path" in result

    @patch("pymilvus.bulk_writer.stage_bulk_writer.Path")
    def test_local_rm(self, mock_path: Mock, stage_bulk_writer: StageBulkWriter) -> None:
        """Test local file removal."""
        # Test successful removal
        mock_file = mock_path.return_value
        mock_file.parent.iterdir.return_value = []

        stage_bulk_writer._local_rm("test_file.parquet")

        mock_file.unlink.assert_called_once()

    @patch.object(StageBulkWriter, "_upload_object")
    @patch.object(StageBulkWriter, "_local_rm")
    @patch("pymilvus.bulk_writer.stage_bulk_writer.Path")
    def test_upload(
        self, mock_path_class: Mock, mock_rm: Mock, mock_upload_object: Mock, stage_bulk_writer: StageBulkWriter
    ) -> None:
        """Test uploading files."""
        # Mock Path behavior
        mock_path = Mock()
        mock_path_class.return_value = mock_path
        mock_path.relative_to.return_value = Path("test.parquet")

        file_list = ["test_file.parquet"]
        result = stage_bulk_writer._upload(file_list)

        assert len(result) == 1
        mock_upload_object.assert_called_once()
        # The actual call will be with the mock path object
        assert mock_rm.called

    @patch.object(StageFileManager, "upload_file_to_stage")
    def test_upload_object(
        self, mock_upload_to_stage: Mock, stage_bulk_writer: StageBulkWriter
    ) -> None:
        """Test uploading a single object."""
        stage_bulk_writer._upload_object("local_file.parquet", "remote_file.parquet")

        stage_bulk_writer._stage_file_manager.upload_file_to_stage.assert_called_once_with(
            "local_file.parquet", stage_bulk_writer._remote_path
        )

    def test_context_manager(self, simple_schema: CollectionSchema) -> None:
        """Test StageBulkWriter as context manager."""
        with patch("pymilvus.bulk_writer.stage_bulk_writer.StageFileManager"), StageBulkWriter(
                schema=simple_schema,
                remote_path="test/data",
                cloud_endpoint="https://api.cloud.zilliz.com",
                api_key="test_api_key",
                stage_name="test_stage",
            ) as writer:
                assert writer is not None
                writer.append_row({
                    "id": 1,
                    "vector": [1.0] * 128,
                    "text": "test",
                })

    @patch.object(StageBulkWriter, "_upload_object")
    @patch("pymilvus.bulk_writer.stage_bulk_writer.Path")
    def test_upload_error_handling(
        self, mock_path_class: Mock, mock_upload_object: Mock, stage_bulk_writer: StageBulkWriter
    ) -> None:
        """Test error handling during upload."""
        mock_upload_object.side_effect = Exception("Upload error")

        # Mock Path behavior
        mock_path = Mock()
        mock_path_class.return_value = mock_path
        mock_path.relative_to.return_value = Path("test.parquet")

        with pytest.raises(MilvusException, match="Failed to upload file"):
            stage_bulk_writer._upload(["test_file.parquet"])


class TestIntegration:
    """Integration tests for stage operations."""

    @pytest.fixture
    def mock_server_responses(self) -> Dict[str, Any]:
        """Mock server responses for integration testing."""
        return {
            "apply_stage": {
                "code": 0,
                "message": "success",
                "data": {
                    "stageName": "test_stage",
                    "stagePrefix": "prefix/",
                    "endpoint": "s3.amazonaws.com",
                    "bucketName": "test-bucket",
                    "region": "us-west-2",
                    "cloud": "aws",
                    "condition": {"maxContentLength": 1073741824},
                    "credentials": {
                        "tmpAK": "test_access_key",
                        "tmpSK": "test_secret_key",
                        "sessionToken": "test_token",
                        "expireTime": "2099-12-31T23:59:59Z",
                    },
                },
            },
            "list_stages": {
                "code": 0,
                "message": "success",
                "data": {"stages": ["stage1", "stage2", "test_stage"]},
            },
        }

    @patch("pymilvus.bulk_writer.stage_restful.requests.post")
    @patch("pymilvus.bulk_writer.stage_restful.requests.get")
    def test_full_stage_workflow(
        self,
        mock_get: Mock,
        mock_post: Mock,
        mock_server_responses: Dict[str, Any],
    ) -> None:
        """Test complete stage workflow from creation to upload."""
        # Setup mock responses
        mock_post_response = Mock()
        mock_post_response.status_code = 200
        mock_post_response.json.return_value = mock_server_responses["apply_stage"]
        mock_post.return_value = mock_post_response

        mock_get_response = Mock()
        mock_get_response.status_code = 200
        mock_get_response.json.return_value = mock_server_responses["list_stages"]
        mock_get.return_value = mock_get_response

        # Create stage manager
        stage_manager = StageManager(
            cloud_endpoint="https://api.cloud.zilliz.com",
            api_key="test_api_key",
        )

        # List stages
        result = stage_manager.list_stages(project_id="test_project")
        assert "test_stage" in result.json()["data"]["stages"]

        # Create stage file manager
        file_manager = StageFileManager(
            cloud_endpoint="https://api.cloud.zilliz.com",
            api_key="test_api_key",
            stage_name="test_stage",
            connect_type=ConnectType.AUTO,
        )

        # Verify stage info can be refreshed
        file_manager._refresh_stage_and_client("data/")
        assert file_manager.stage_info["stageName"] == "test_stage"



from pymilvus import MilvusClient, DataType, Function, FunctionType

# 1. Setup Milvus Client
client = MilvusClient("http://localhost:19530")
COLLECTION_NAME = "multilingual_test_B"
if client.has_collection(collection_name=COLLECTION_NAME):
    client.drop_collection(collection_name=COLLECTION_NAME)

# 2. Define analyzers for multiple languages
# These individual analyzer definitions will be reused by both methods.
analyzers = {
    "Japanese": { 
        # Use lindera with japanese dict 'ipadic' 
        # and remove punctuation beacuse lindera tokenizer will remain punctuation
        "tokenizer":{
            "type": "lindera",
            "dict_kind": "ipadic"
        },
        "filter": ["removepunct"]
    },
    "English": {
        # Use build-in english analyzer
        "type": "english",
    },
    "default": {
        # use icu tokenizer as a fallback.
        "tokenizer": "icu",
    }
}

# --- Option B: Using Language Identifier Tokenizer ---
print("\n--- Demonstrating Language Identifier Tokenizer ---")

# 3B. reate a collection with language identifier
analyzer_params_langid = {
    "tokenizer": {
        "type": "language_identifier",
        "analyzers": analyzers # Referencing the analyzers defined in Step 2
    },
}

schema_langid = MilvusClient.create_schema(
    auto_id=True,
    enable_dynamic_field=False,
)
schema_langid.add_field(field_name="id", datatype=DataType.INT64, is_primary=True)
# The 'language' field is not strictly needed by the analyzer itself here, as detection is automatic.# However, you might keep it for metadata purposes.
schema_langid.add_field(field_name="text", datatype=DataType.VARCHAR, max_length=25565, enable_analyzer=True, analyzer_params = analyzer_params_langid)
schema_langid.add_field(field_name="text_sparse", datatype=DataType.SPARSE_FLOAT_VECTOR) # BM25 Sparse Vector# add bm25 function
text_bm25_function_langid = Function(
    name="text_bm25",
    function_type=FunctionType.BM25,
    input_field_names=["text"],
    output_field_names=["text_sparse"],
)
schema_langid.add_function(text_bm25_function_langid)

index_params_langid = client.prepare_index_params()
index_params_langid.add_index(
    field_name="text_sparse",
    index_type="AUTOINDEX", # Use auto index for BM25
    metric_type="BM25",
)

client.create_collection(
    collection_name=COLLECTION_NAME,
    schema=schema_langid,
    index_params=index_params_langid
)
print(f"Collection '{COLLECTION_NAME}' created successfully with Language Identifier Tokenizer.")

# 4B. Insert Data for Language Identifier Tokenizer and Load Collection
# Insert English and Japanese movie titles. The language_identifier will detect the language.
client.insert(
    collection_name=COLLECTION_NAME,
    data=[
        {"text": "The Lord of the Rings"},
        {"text": "Spirited Away"},
        {"text": "千と千尋の神隠し"}, 
        {"text": "君の名は。"},
    ]
)
print(f"Inserted multilingual data into '{COLLECTION_NAME}'.")

# Load the collection into memory before searching
client.load_collection(collection_name=COLLECTION_NAME)

# 5B. Perform a full-text search with Language Identifier Tokenizer# No need to specify analyzer_name in search_params; it's detected automatically for the query.
print("\n--- Search results for Language Identifier Tokenizer ---")
results_langid_jp = client.search(
    collection_name=COLLECTION_NAME,
    data=["神隠し"],
    limit=2,
    output_fields=["text"],
    search_params={"metric_type": "BM25"}, # Analyzer automatically determined by language_identifier
    consistency_level = "Strong",
)
print("\nSearch results for '神隠し' (Language Identifier Tokenizer):")
for result in results_langid_jp[0]:
    print(result)

results_langid_en = client.search(
    collection_name=COLLECTION_NAME,
    data=["the Rings"],
    limit=2,
    output_fields=["text"],
    search_params={"metric_type": "BM25"}, # Analyzer automatically determined by language_identifier
    consistency_level = "Strong",
)
print("\nSearch results for 'the Rings' (Language Identifier Tokenizer):")
for result in results_langid_en[0]:
    print(result)

client.drop_collection(collection_name=COLLECTION_NAME)
print(f"Collection '{COLLECTION_NAME}' dropped.")